"""Cache class for key/value pairs"""

import time
from collections import OrderedDict
import pickle
import atexit

# Local imports
from zat.utils import file_storage


class Cache(object):
    """In process memory cache. Not thread safe.
       Usage:
            cache = Cache(max_size=5, timeout=10)
            cache.set('foo', 'bar')
            cache.get('foo')
            >>> bar
            time.sleep(11)
            cache.get('foo')
            >>> None
            cache.clear()
    """
    def __init__(self, max_size=1000, timeout=None, load=None):
        """Cache Initialization"""
        self.disk_storage = file_storage.FileStorage()
        self.load = load
        storage_bytes = None
        if load:
            storage_bytes = self.disk_storage.get(load)  # This can fail, returning None
        self.store = pickle.loads(storage_bytes) if storage_bytes else OrderedDict()
        self.max_size = max_size
        self.timeout = timeout
        self._compression_timer = 60
        self._last_compression = time.time()

        # Try to do cleanup/serialization at exit
        atexit.register(self.cleanup)

    def set(self, key, value):
        """Add an item to the cache
        Args:
               key: item key
               value: the value associated with this key
        """
        self._check_limit()
        _expire = time.time() + self.timeout if self.timeout else None
        self.store[key] = (value, _expire)

    def get(self, key):
        """Get an item from the cache
           Args:
               key: item key
           Returns:
               the value of the item or None if the item isn't in the cache
        """
        data = self.store.get(key)
        if not data:
            return None
        value, expire = data
        if expire and time.time() > expire:
            del self.store[key]
            return None
        return value

    def clear(self):
        """Clear the cache"""
        self.store = OrderedDict()

    def dump(self):
        """Dump the cache (for debugging)"""
        for key in self.store.keys():
            print(key, ':', self.get(key))

    @property
    def size(self):
        return len(self.store)

    def cleanup(self):
        print('Calling cleanup...')
        self.persist()

    def persist(self):
        """Serialize the cache to disk"""
        if self.load:
            store_bytes = pickle.dumps(self.store)
            self.disk_storage.store(self.load, store_bytes)

    def _check_limit(self):
        """Intenal method: check if current cache size exceeds maximum cache
           size and pop the oldest item in this case"""

        # First compress
        self._compress()

        # Then check the max size
        if len(self.store) >= self.max_size:
            self.store.popitem(last=False)  # FIFO

    def _compress(self):
        """Internal method to compress the cache. This method will
           expire any old items in the cache, making the cache smaller"""

        # Don't compress too often
        now = time.time()
        if self._last_compression + self._compression_timer < now:
            self._last_compression = now
            for key in list(self.store.keys()):
                self.get(key)


def test():
    """Test for the Cache class"""

    # Create the Cache
    my_cache = Cache(max_size=5, timeout=1)
    my_cache.set('foo', 'bar')

    # Test storage
    assert my_cache.get('foo') == 'bar'

    # Test timeout
    time.sleep(1.1)
    assert my_cache.get('foo') is None

    # Test max_size
    my_cache = Cache(max_size=5)
    for i in range(6):
        my_cache.set(str(i), i)

    # So the '0' key should no longer be there FIFO
    assert my_cache.get('0') is None
    assert my_cache.get('5') is not None

    # Make sure size is working
    assert my_cache.size == 5

    # Dump the cache
    my_cache.dump()

    # Test storing 'null' values
    my_cache.set(0, 'foo')
    my_cache.set(0, 'bar')
    my_cache.set(None, 'foo')
    my_cache.set('', None)
    assert my_cache.get('') is None
    assert my_cache.get(None) == 'foo'
    assert my_cache.get(0) == 'bar'

    # Test the cache compression
    my_cache = Cache(max_size=5, timeout=1)
    for i in range(5):
        my_cache.set(str(i), i)
    my_cache._compression_timer = 1
    assert my_cache.size == 5

    # Make sure compression is working
    time.sleep(1.1)
    my_cache._compress()
    assert my_cache.size == 0

    # Also make sure compression call is throttled
    my_cache._compress()  # Should not output a compression message

    # Test persistance functionality
    my_cache = Cache(load='my_test_cache')
    for i in range(5):
        my_cache.set(str(i), i)
    assert my_cache.size == 5

    my_cache.persist()
    del my_cache

    load_cache = Cache(load='my_test_cache')
    assert load_cache.size == 5
    load_cache.dump()


if __name__ == '__main__':

    # Run the test
    test()



"""HTTP User Agents Example"""

import os
import sys
import argparse
from collections import Counter
from pprint import pprint

# Local imports
from zat import zeek_log_reader

if __name__ == '__main__':
    # Example to run the zeek log reader on a given file

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('zeek_log', type=str, help='Specify a zeek log to run ZeekLogReader test on')
    parser.add_argument('-t', action='store_true', default=False, help='Sets the program to tail a live Zeek log')
    parser.add_argument('-s', action='store_true', default=False, help='Only print the summary of the findings.')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Sanity check that this is a http log
    if 'http' not in args.zeek_log:
        print('This example only works with Zeek http.log files..')
        sys.exit(1)

    # File may have a tilde in it
    if args.zeek_log:
        args.zeek_log = os.path.expanduser(args.zeek_log)

        # Run the zeek reader on a given log file counting up user agents
        http_agents = Counter()
        reader = zeek_log_reader.ZeekLogReader(args.zeek_log, tail=args.t)
        for count, row in enumerate(reader.readrows()):
            # Track count
            http_agents[row['user_agent']] += 1

            # Every 10000 rows report agent counts (least common)
            if not args.s:
                if count % 10000 == 0:
                    print('\n<<<Least Common User Agents>>>')
                    pprint(http_agents.most_common()[:-50:-1])

        # Also report at the end (if there is one)
        print('\nMost Common User Agents:')
        pprint(http_agents.most_common()[:20])

        print('\nLeast Common User Agents:')
        pprint(http_agents.most_common()[:-50:-1])



"""LogToDataFrame: Converts a Zeek log to a Pandas DataFrame"""


# Third Party
import pandas as pd

# Local
from zat import zeek_log_reader


class LogToDataFrame(object):
    """LogToDataFrame: Converts a Zeek log to a Pandas DataFrame
        Notes:
            This class has recently been overhauled from a simple loader to a more
            complex class that should in theory:
              - Select better types for each column
              - Should be faster
              - Produce smaller memory footprint dataframes
            If you have any issues/problems with this class please submit a GitHub issue.
        More Info: https://supercowpowers.github.io/zat/large_dataframes.html
    """
    def __init__(self):
        """Initialize the LogToDataFrame class"""

        # First Level Type Mapping
        #    This map defines the types used when first reading in the Zeek log into a 'chunk' dataframes.
        #    Types (like time and interval) will be defined as one type at first but then
        #    will undergo further processing to produce correct types with correct values.
        # See: https://stackoverflow.com/questions/29245848/what-are-all-the-dtypes-that-pandas-recognizes
        #      for more info on supported types.
        self.type_map = {'bool': 'category',  # Can't hold NaN values in 'bool', so we're going to use category
                         'count': 'UInt64',
                         'int': 'Int32',
                         'double': 'float',
                         'time': 'float',      # Secondary Processing into datetime
                         'interval': 'float',  # Secondary processing into timedelta
                         'port': 'UInt16'
                         }

    def _get_field_info(self, log_filename):
        """Internal Method: Use ZAT log reader to read header for names and types"""
        _zeek_reader = zeek_log_reader.ZeekLogReader(log_filename)
        _, field_names, field_types, _ = _zeek_reader._parse_zeek_header(log_filename)
        return field_names, field_types

    def _create_initial_df(self, log_filename, all_fields, usecols, dtypes):
        """Internal Method: Create the initial dataframes by using Pandas read CSV (primary types correct)"""
        return pd.read_csv(log_filename, sep='\t', names=all_fields, usecols=usecols, dtype=dtypes, comment="#", na_values='-')

    def create_dataframe(self, log_filename, ts_index=True, aggressive_category=True, usecols=None):
        """ Create a Pandas dataframe from a Bro/Zeek log file
            Args:
               log_fllename (string): The full path to the Zeek log
               ts_index (bool): Set the index to the 'ts' field (default = True)
               aggressive_category (bool): convert unknown columns to category (default = True)
               usecol (list): A subset of columns to read in (minimizes memory usage) (default = None)
        """

        # Grab the field information
        field_names, field_types = self._get_field_info(log_filename)
        all_fields = field_names  # We need ALL the fields for later

        # If usecols is set then we'll subset the fields and types
        if usecols:
            # Usecols needs to include ts
            if 'ts' not in usecols:
                usecols.append('ts')
            field_types = [t for t, field in zip(field_types, field_names) if field in usecols]
            field_names = [field for field in field_names if field in usecols]

        # Get the appropriate types for the Pandas Dataframe
        pandas_types = self.pd_column_types(field_names, field_types, aggressive_category)

        # Now actually read in the initial dataframe
        self._df = self._create_initial_df(log_filename, all_fields, usecols, pandas_types)

        # Now we convert 'time' and 'interval' fields to datetime and timedelta respectively
        for name, zeek_type in zip(field_names, field_types):
            if zeek_type == 'time':
                self._df[name] = pd.to_datetime(self._df[name], unit='s')
            if zeek_type == 'interval':
                self._df[name] = pd.to_timedelta(self._df[name], unit='s')

        # Set the index
        if ts_index and not self._df.empty:
            self._df.set_index('ts', inplace=True)
        return self._df

    def pd_column_types(self, column_names, column_types, aggressive_category=True, verbose=False):
        """Given a set of names and types, construct a dictionary to be used
           as the Pandas read_csv dtypes argument"""

        # Agressive Category means that types not in the current type_map are
        # mapped to a 'category' if aggressive_category is False then they
        # are mapped to an 'object' type
        unknown_type = 'category' if aggressive_category else 'object'

        pandas_types = {}
        for name, zeek_type in zip(column_names, column_types):

            # Grab the type
            item_type = self.type_map.get(zeek_type)

            # Sanity Check
            if not item_type:
                # UID/FUID/GUID always gets mapped to object
                if 'uid' in name:
                    item_type = 'object'
                else:
                    if verbose:
                        print('Could not find type for {:s} using {:s}...'.format(zeek_type, unknown_type))
                    item_type = unknown_type

            # Set the pandas type
            pandas_types[name] = item_type

        # Return the dictionary of name: type
        return pandas_types


# Simple test of the functionality
def test():
    """Test for LogToDataFrame Class"""
    import os
    pd.set_option('display.width', 1000)
    from zat.utils import file_utils

    # Grab a test file
    data_path = file_utils.relative_dir(__file__, '../data')
    log_path = os.path.join(data_path, 'conn.log')

    # Convert it to a Pandas DataFrame
    log_to_df = LogToDataFrame()
    my_df = log_to_df.create_dataframe(log_path)

    # Print out the head
    print(my_df.head())

    # Print out the datatypes
    print(my_df.dtypes)

    # Test a bunch
    tests = ['app_stats.log', 'dns.log', 'http.log', 'notice.log', 'tor_ssl.log',
             'conn.log', 'dhcp_002.log', 'files.log',  'smtp.log', 'weird.log',
             'ftp.log',  'ssl.log', 'x509.log']
    for log_path in [os.path.join(data_path, log) for log in tests]:
        print('Testing: {:s}...'.format(log_path))
        my_df = log_to_df.create_dataframe(log_path)
        print(my_df.head())
        print(my_df.dtypes)

    # Test out usecols arg
    conn_path = os.path.join(data_path, 'conn.log')
    my_df = log_to_df.create_dataframe(conn_path, usecols=['id.orig_h', 'id.orig_p', 'id.resp_h', 'id.resp_p',
                                                           'proto', 'orig_bytes', 'resp_bytes'])

    # Test an empty log (a log with header/close but no data rows)
    log_path = os.path.join(data_path, 'http_empty.log')
    my_df = log_to_df.create_dataframe(log_path)

    # Print out the head
    print(my_df.head())

    # Print out the datatypes
    print(my_df.dtypes)

    print('LogToDataFrame Test successful!')


if __name__ == '__main__':
    # Run the test for easy testing/debugging
    test()



"""Zeek log to Parquet Dataframe Example"""
import os
import sys
import argparse

# Note: We're going to import pyarrow but it currently has an open issue around supporting time deltas
#    - https://issues.apache.org/jira/browse/ARROW-6780 so we have to convert timedelta fields to str
# Also see:
#   - https://stackoverflow.com/questions/53893554/transfer-and-write-parquet-with-python-and-pandas-got-timestamp-error
#
from datetime import timedelta
import pandas as pd
try:
    import pyarrow
except ImportError:
    print('Please > pip install pyarrow')
    sys.exit(1)

# Local imports
from zat.log_to_dataframe import LogToDataFrame


# Helper method for temporarily converting timedelta to string
def convert_timedelta_to_str(df):
    delta_columns = df.select_dtypes(include=['timedelta'])
    for column in delta_columns:
        df[column] = df[column].apply(tdelta_value_to_str)
    return df


def tdelta_value_to_str(value):
    if pd.isnull(value):
        return '-'  # Standard for Zeek null value
    else:
        return str(timedelta(seconds=value.total_seconds()))


if __name__ == '__main__':
    # Example to write Parquet file from a zeek log

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('zeek_log', type=str, help='Specify the zeek log input file')
    parser.add_argument('parquet_file', type=str, help='Specify the parquet file to write out')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # File may have a tilde in it
    if args.zeek_log and args.parquet_file:
        args.zeek_log = os.path.expanduser(args.zeek_log)
        args.parquet_file = os.path.expanduser(args.parquet_file)

        # Convert to dataframe and write out the parquet file
        log_to_df = LogToDataFrame()
        zeek_df = log_to_df.create_dataframe(args.zeek_log)
        print('Dataframe Created: {:d} rows...'.format(len(zeek_df)))

        # Check for any timedelta fields (see note above)
        df = convert_timedelta_to_str(zeek_df)

        zeek_df.to_parquet(args.parquet_file, compression='snappy', use_deprecated_int96_timestamps=True)
        print('Complete: {:s} --> {:s}'.format(args.zeek_log, args.parquet_file))



"""Run a set of Yara Rule matches on Extracted Files
           Note: Download yara rules from their repo and give index file as arg
           $ git clone https://github.com/Yara-Rules/rules rules
           $ python yara_matches -r /path/to/rules/index.yar -e /path/to/zeek/extract_files
"""

import os
import sys
import time
import argparse
from pprint import pprint

# Third Party Imports
try:
    import yara
except ImportError:
    print('\nThis example needs yara. Please do a $pip install yara-python')
    sys.exit(1)

# Local imports
from zat.utils import dir_watcher, signal_utils

def yara_match(file_path, rules):
    """Callback for a newly extracted file"""
    print('New Extracted File: {:s}'.format(file_path))
    print('Mathes:')
    pprint(rules.match(file_path))

def my_exit():
    """Exit on Signal"""
    print('Goodbye...')
    sys.exit()

if __name__ == '__main__':
    # Run a set of Yara Rule matches on Extracted Files

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('-r', '--rule-index', type=str, required=True, help='Specify the yara rule index file (e.g. /full/path/to/yara/rules/index.yar)')
    parser.add_argument('-e', '--extract-dir', type=str, required=True, help='Specify the Zeek extract_files directory (e.g. /full/path/to/zeek/extract_files)')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # If no args just call help
    if len(sys.argv) == 1:
        parser.print_help()
        print('\nNote: Download the yara repo and give the index file as an arg')
        print('$ git clone https://github.com/Yara-Rules/rules')
        print('$ python yara_matches -r /path/to/rules/index.yar -e /path/to/zeek/extract_files')
        sys.exit(1)

    # Sanity check that the args exist and are what we expect
    if not os.path.isfile(args.rule_index):
        print('--rule-index file not found.. should be /full/path/to/yara/rules/index.yar')
        sys.exit(1)
    if not os.path.isdir(args.extract_dir):
        print('--extract-dir directory not found.. should be /full/path/to/zeek/extract_files')
        sys.exit(1)

    # Load/compile the yara rules
    my_rules = yara.compile(args.rule_index)

    # Create DirWatcher and start watching the Zeek extract_files directory
    print('Watching Extract Files Directory: {:s}'.format(args.extract_dir))
    dir_watcher.DirWatcher(args.extract_dir, callback=yara_match, rules=my_rules)

    # Okay so just wait around for files to be dropped by Zeek or someone hits Ctrl-C
    with signal_utils.signal_catcher(my_exit):
        while True:
            time.sleep(.5)



"""GeoLookup Class"""


import time
import pprint

# Third Party
import requests

# Local imports
from zat.utils import cache


class GeoLookup(object):
    """GeoLookup: Query IPStack API (https://ipstack.com/) for Geographic information,
                 the class contains a Cache to minimize queries.

        Args:
            apikey (str): The API key to use for IPStack queries (default=None)
            summary (bool): Just return summary information for GeoLookup (default=True)
            max_cache_size (int): Maximum size of query cache (default=10000)
            max_cache_time (int): Time to keep query results in cache (default=30 days)
    """

    def __init__(self, apikey=None, summary=True, max_cache_size=10000, max_cache_time=30, throttle=True):
        """GeoLookup Init"""

        # Public API Key
        # Note: The key below is a low-volume public key. Please call this method with your own API key :)
        if apikey is None:
            print('Using public API Key: Please set apikey=<your key> when creating this class')
        pub_apikey = '7278776fac2f2c66b1c760c650c20e8e'
        self.known_ipstack_ips = ['158.85.167.221', '158.85.167.217', '158.85.167.201']
        self.apikey = apikey or pub_apikey
        self.summary = summary
        self.throttle = throttle

        # Create query cache
        seconds = max_cache_time*24*60*60  # Convert from days
        self.query_cache = cache.Cache(max_size=max_cache_size, timeout=seconds, load='zat_geo_cache')  # Convert to Seconds

    @property
    def size(self):
        return self.query_cache.size

    def query_ip(self, ip_address, verbose=False):
        """Query method for the IpStack Service
            Args:
               ip_address(str): The IP Address to be queried
        """

        # Is this a known IpStack ip address?
        if ip_address in self.known_ipstack_ips:
            return None

        # First check query cache
        cached = self.query_cache.get(ip_address)
        if cached:
            if verbose:
                print('Returning Cached Query Results')
            return cached

        # Not in cache so make the actual query
        url = 'http://api.ipstack.com/' + ip_address
        response = requests.post(url, params={'access_key': self.apikey})

        # Make sure we got a json blob back
        try:
            output = response.json()
        except ValueError:
            error_msg = 'No valid response, throttling and trying again...'
            if self.throttle:
                if verbose:
                    print(error_msg)
                time.sleep(30)
                return self.query_ip(ip_address)

            return {'error': error_msg}

        # Check for error or not-found
        if output['type'] is None:
            output = None
            self.query_cache.set(ip_address, output)
            return output

        # Summary removes 'location' info (flag, calling_code, languages, etc)
        if self.summary:
            del output['location']

        # Put results in Cache
        self.query_cache.set(ip_address, output)
        self.query_cache.persist()

        # Return results
        return output


# Unit test: Create the class and test it
def test():
    """geo_lookup.py test"""

    # Execute the worker (unit test)
    geo_lookup = GeoLookup(summary=False)
    output = geo_lookup.query_ip('73.26.145.66')
    print('\n<<< Unit Test FULL>>>')
    pprint.pprint(output)

    geo_lookup = GeoLookup()
    output = geo_lookup.query_ip('73.26.145.66')
    print('\n<<< Unit Test Summary>>>')
    pprint.pprint(output)
    output = geo_lookup.query_ip('123.4.5.6')
    print('\n<<< Unit Test Summary>>>')
    pprint.pprint(output)

    # Test Cache
    output = geo_lookup.query_ip('73.26.145.66')
    print('\n<<< Unit Test Cache>>>')
    pprint.pprint(output)

    # Test Size
    assert geo_lookup.size == 2

    # Test some error conditions
    output = geo_lookup.query_ip('123')
    print('\n<<< Unit Test Malformed IP Query>>>')
    pprint.pprint(output)


if __name__ == "__main__":
    test()



"""LiveSimulator: This class reads in various Zeek logs. The class utilizes
                 the ZeekLogReader and simply loops over the static zeek log
                 file, replaying rows and changing any time stamps
        Args:
            eps (int): Events Per Second that the simulator will emit events (default = 10)
            max_rows (int): The maximum number of rows to generate (default = None (go forever))
"""

import os
import time
import datetime
import itertools

# Third party
import numpy as np

# Local Imports
from zat import zeek_log_reader
from zat.utils import file_utils


class LiveSimulator(object):
    """LiveSimulator: This class reads in various Zeek logs. The class utilizes the
                      ZeekLogReader and simply loops over the static zeek log file
                      replaying rows at the specified EPS and changing timestamps to 'now()'
    """

    def __init__(self, filepath, eps=10, max_rows=None, only_once=False):
        """Initialization for the LiveSimulator Class
           Args:
               eps (int): Events Per Second that the simulator will emit events (default = 10)
               max_rows (int): The maximum number of rows to generate (default = None (go forever))
        """

        # Compute EPS timer
        # Logic:
        #     - Normal distribution centered around 1.0/eps
        #     - Make sure never less than 0
        #     - Precompute 1000 deltas and then just cycle around
        self.eps_timer = itertools.cycle([max(0, delta) for delta in np.random.normal(1.0/float(eps), .5/float(eps), size=1000)])

        # Initialize the Zeek log reader
        self.log_reader = zeek_log_reader.ZeekLogReader(filepath, tail=False)

        # Store max_rows and only_once flag
        self.max_rows = max_rows
        self.only_once = only_once

    def rows(self):
        """Using the ZeekLogReader this method generates (yields) each row of the log file
           replacing timestamps, looping and emitting rows based on EPS rate
        """

        # Loop forever or until max_rows is reached
        num_rows = 0
        while True:

            # Yield the rows from the internal reader
            for row in self.log_reader.readrows():
                yield self.replace_timestamp(row)

                # Sleep and count rows
                time.sleep(next(self.eps_timer))
                num_rows += 1

                # Check for max_rows
                if self.max_rows and (num_rows >= self.max_rows):
                    return

            # Check for only_once
            if self.only_once:
                return

    @staticmethod
    def replace_timestamp(row):
        """Replace the timestamp with now()"""
        if 'ts' in row:
            row['ts'] = datetime.datetime.utcnow()
        return row


def test():
    """Test for LiveSimulator Python Class"""

    # Grab a test file
    data_path = file_utils.relative_dir(__file__, '../data')
    test_path = os.path.join(data_path, 'conn.log')
    print('Opening Data File: {:s}'.format(test_path))

    # Create a LiveSimulator reader
    data_stream = LiveSimulator(test_path, max_rows=10)
    for line in data_stream.rows():
        print(line)
    print('Read with max_rows Test successful!')


if __name__ == '__main__':
    # Run the test for easy testing/debugging
    test()



"""DataFrameToMatrix: Class that converts a DataFrame to a Numpy Matrix (ndarray)"""


# Third Party
import pandas as pd
import numpy as np


class DataFrameToMatrix():
    """DataFrameToMatrix: Class that converts a DataFrame to a Numpy Matrix (ndarray)
        Notes:
            fit_transform: Does a fit and a transform and returns the transformed matrix
            transform: Based on previous fit parameters returns the transformed matrix
        More Info: https://supercowpowers.github.io/zat/dataframe_to_matrix.html

        # Nullable integer arrays are currently not handled by Numpy
        # Cast Nullable integer arrays to float32
        null_int_types = [pd.UInt16Dtype, pd.UInt32Dtype, pd.UInt64Dtype, pd.Int64Dtype]
        for col in _internal_df:
            if type(_internal_df[col].dtype) in null_int_types:
                _internal_df[col] = _internal_df[col].astype(np.float32)

    """
    def __init__(self):
        """Initialize the DataFrameToMatrix class"""
        self.column_names = None
        self.norm_map = {}
        self.dtype_map = {}
        self.nan_replace = -999

    def fit_transform(self, input_df, normalize=True, nan_replace=-999, copy=True):
        """Convert the dataframe to a matrix (numpy ndarray)
        Args:
            input_df (dataframe): The dataframe to convert
            normalize (bool): Boolean flag to normalize numeric columns (default=True)
        """
        self.nan_replace = nan_replace

        # Copy the dataframe (if wanted)
        _internal_df = input_df.copy() if copy else input_df

        # Convert object columns to categorical
        self.object_to_categorical(_internal_df)

        # Convert categorical NaNs
        self.fit_category_nans(_internal_df)

        # Lock categories to specific values (important for train/predict consistency)
        self.lock_categorical(_internal_df)

        # Sanity Check
        self.sanity_check_categorical(_internal_df)

        # Normalize numeric columns (mean normalize, sometimes called 'standardizing')
        if normalize:
            self.normalize_numeric(_internal_df)

        # Remove any numerical NaNs (categorical NaNs were removed above)
        for column in _internal_df.select_dtypes(include='number').columns:
            _internal_df[column].fillna(self.nan_replace, inplace=True)

        # Drop any columns that aren't numeric or categorical
        for column in list(_internal_df.select_dtypes(exclude=['number', 'category']).columns):
            print('Dropping {:s} column...'.format(column))
        _internal_df = _internal_df.select_dtypes(include=['number', 'category'])

        # Capture all the column/dtype information from the dataframe
        self.column_names = _internal_df.columns.to_list()
        for column in _internal_df.columns:
            self.dtype_map[column] = _internal_df[column].dtype

        # Now with every thing setup, call the dummy_encoder, convert to ndarray and return
        return pd.get_dummies(_internal_df).to_numpy(dtype=np.float32)

    def transform(self, input_df, copy=True):
        """Convert the dataframe to a matrix (numpy ndarray)
        Args:
            input_df (dataframe): The dataframe to convert
        """

        # Make sure we have the same columns in the right order
        # Copy the dataframe (if wanted)
        _internal_df = input_df[self.column_names].copy() if copy else input_df[self.column_names]

        # Convert all columns to the proper dtypes
        for column, dtype in self.dtype_map.items():
            _internal_df[column] = _internal_df[column].astype(dtype)

        # Convert any categorical NaNs to a 'NaN' category
        self.transform_category_nans(_internal_df)

        # Normalize any numeric columns
        for column, (smin, smax) in self.norm_map.items():
            print('Normalizing column {:s}...'.format(column))
            _internal_df[column] = (_internal_df[column] - smin) / (smax - smin)

        # Remove any numerical NaNs (categorical NaNs were removed above)
        for column in _internal_df.select_dtypes(include='number').columns:
            _internal_df[column].fillna(self.nan_replace, inplace=True)

        # Now with every thing setup, call the dummy_encoder, convert to ndarray and return
        return pd.get_dummies(_internal_df).to_numpy(dtype=np.float32)

    @staticmethod
    def fit_category_nans(df):
        """ONLY FIT: Convert np.NaNs to a category 'NaN'"""
        for column in df.select_dtypes(include=['category']).columns:
            if df[column].isnull().any():
                df[column].cat.add_categories('NaN', inplace=True)
                df[column].fillna('NaN', inplace=True)

    @staticmethod
    def transform_category_nans(df):
        """ONLY TRANSFORM: Convert np.NaNs to a category 'NaN'"""
        for column in df.select_dtypes(include=['category']).columns:
            if 'NaN' in df[column].cat.categories:
                df[column].fillna('NaN', inplace=True)

    @staticmethod
    def object_to_categorical(df):
        """Run a heuristic on the object columns to determine whether it contains categorical values
           if the heuristic decides it's categorical then the type of the column is changed
        Args:
            df (dataframe): The dataframe to check for categorical data
        Returns:
            None but the dataframe columns are modified
        """

        # Loop through each column that might be converable to categorical
        for column in df.select_dtypes(include='object').columns:

            # If we don't have too many unique values convert the column
            if df[column].nunique() < 100:
                print('Changing column {:s} to category...'.format(column))
                df[column] = pd.Categorical(df[column])

    @staticmethod
    def lock_categorical(df):
        """Lock the categorical column types to a specific ordered list of categories
        Args:
            df (dataframe): The dataframe to lock categorical columns
        Returns:
            None but note that the dataframe is modified to 'lock' the categorical columns
        """
        for column in df.select_dtypes(include='category').columns:
            df[column] = pd.Categorical(df[column], categories=sorted(df[column].unique().tolist()))

    @staticmethod
    def sanity_check_categorical(df):
        """Sanity check for 'dimensionality explosion' on categorical types
        Args:
            df (dataframe): The dataframe to check the categorical columns
        Returns:
            None
        """
        for column in df.select_dtypes(include='category').columns:
            # Give warning on category types will LOTs of values
            num_unique = df[column].nunique()
            if num_unique > 20:
                print('WARNING: {:s} will expand into {:d} dimensions...'.format(column, num_unique))

    def normalize_numeric(self, df):
        """Normalize (mean normalize) the numeric columns in the dataframe
        Args:
            df (dataframe): The dataframe to normalize
        Returns:
            None but note that the numeric columns of the dataframe are modified
        """
        for column in df.select_dtypes(include='number').columns:
            print('Normalizing column {:s}...'.format(column))
            df[column] = self._normalize_series(df[column])

    def _normalize_series(self, series):
        smin = series.min()
        smax = series.max()

        # Check for div by 0
        if smax - smin == 0:
            print('Cannot normalize series (div by 0) so not normalizing...')
            return series

        # Capture the normalization info and return the normalize series
        self.norm_map[series.name] = (smin, smax)
        return (series - smin) / (smax - smin)


# Simple test of the functionality
def test():
    """Test for DataFrameToMatrix Class"""
    import os
    import pickle
    from tempfile import NamedTemporaryFile
    import numpy.testing as np_test_utils
    pd.set_option('display.width', 1000)

    test_df = pd.DataFrame(
        {'A': pd.Categorical(['a', 'b', 'c', 'a'], ordered=True),
         'B': pd.Categorical(['a', 'b', 'c', 'a'], ordered=False),
         'C': pd.Categorical(['a', 'b', 'z', 'a'], categories=['a', 'b', 'z', 'd']),
         'D': [1, 2, 3, 4],
         'E': ['w', 'x', 'y', 'z'],
         'F': [1.1, 2.2, 3.3, 4.4],
         'G': pd.to_datetime([0, 1, 2, 3]),
         'H': [True, False, False, True]
         }
    )
    test_df2 = pd.DataFrame(
        {'A': pd.Categorical(['a', 'b', 'b', 'a'], ordered=True),
         'B': pd.Categorical(['a', 'b', 'd', 'a'], ordered=False),
         'C': pd.Categorical(['a', 'b', 'z', 'y'], categories=['a', 'b', 'z', 'd']),
         'D': [1, 2, 3, 7],
         'E': ['w', 'x', 'z', 'foo'],
         'F': [1.1, 2.2, 3.3, 4.4],
         'H': [True, False, False, False]
         }
    )

    # Copy the test_df for testing later
    copy_test_df = test_df.copy()

    # Test the transformation from dataframe to numpy ndarray and back again
    to_matrix = DataFrameToMatrix()
    print('FIT-TRANSFORM')
    matrix = to_matrix.fit_transform(test_df)
    print('TRANSFORM')
    matrix_test = to_matrix.transform(test_df)

    # These two matrices should be the same
    np_test_utils.assert_equal(matrix, matrix_test)

    # Assert that the dataframe we passed in didn't change
    copy_test_df.equals(test_df)

    # Test that the conversion gives us the same columns on a df with different category values
    # This also tests NaN in a category column
    print('TRANSFORM2')
    matrix2 = to_matrix.transform(test_df2)
    assert matrix.shape == matrix2.shape

    # First two ROWS should be the same
    np_test_utils.assert_equal(matrix[0], matrix2[0])
    np_test_utils.assert_equal(matrix[1], matrix2[1])

    # Test normalize
    to_matrix_norm = DataFrameToMatrix()
    print('FIT-TRANSFORM')
    norm_matrix = to_matrix_norm.fit_transform(test_df)
    print(norm_matrix)
    assert(norm_matrix[:, 0].min() == 0)
    assert(norm_matrix[:, 0].max() == 1)

    # Make sure normalize 'does the right thing' when doing transform
    print('TRANSFORM')
    norm_matrix2 = to_matrix_norm.transform(test_df2)
    assert(norm_matrix2[:, 0].min() == 0)
    assert(norm_matrix2[:, 0].max() == 2)    # Normalization is based on FIT range

    # Test div by zero in normalize
    test_df3 = test_df2.copy()
    test_df3['D'] = [1, 1, 1, 1]
    print('FIT-TRANSFORM')
    norm_matrix3 = to_matrix_norm.fit_transform(test_df3)
    assert(norm_matrix3[:, 0].min() == 1)
    assert(norm_matrix3[:, 0].max() == 1)

    # Test serialization
    temp = NamedTemporaryFile(delete=False)
    pickle.dump(to_matrix, temp)
    temp.close()

    # Deserialize and test
    to_matrix_from_disk = pickle.load(open(temp.name, 'rb'))
    print('TRANSFORM')
    matrix3 = to_matrix_from_disk.transform(test_df)
    print('TRANSFORM')
    matrix4 = to_matrix_from_disk.transform(test_df2)
    np_test_utils.assert_equal(matrix, matrix3)
    np_test_utils.assert_equal(matrix2, matrix4)

    # Remove the temporary file
    os.unlink(temp.name)

    # Try 'nullable' integer arrays
    null_df = test_df2.copy()
    null_df['I'] = pd.Series([10, 11, 12, np.NaN], dtype='UInt64')
    print('FIT-TRANSFORM')
    matrix = to_matrix.fit_transform(null_df)
    print('TRANSFORM')
    matrix_test = to_matrix.transform(null_df)

    # These two matrices should be the same
    np_test_utils.assert_equal(matrix, matrix_test)

    # Now actually try the matrix with a scikit-learn algo
    from sklearn.cluster import KMeans
    to_matrix = DataFrameToMatrix()
    my_matrix = to_matrix.fit_transform(test_df)
    kmeans = KMeans(n_clusters=2).fit_predict(my_matrix)

    # Now we can put our ML results back onto our dataframe!
    test_df['cluster'] = kmeans
    cluster_groups = test_df.groupby('cluster')

    # Now print out the details for each cluster
    for _key, group in cluster_groups:
        print('Rows in Cluster: {:d}'.format(len(group)))
        print(group.head(), '\n')
    del test_df['cluster']

    # Now we're going to intentionally introduce NaNs in the categorical output just to see what happens
    to_matrix = DataFrameToMatrix()
    _ = to_matrix.fit_transform(test_df)
    my_matrix2 = to_matrix.transform(test_df2)
    kmeans = KMeans(n_clusters=2).fit_predict(my_matrix2)

    # Now we can put our ML results back onto our dataframe!
    test_df2['cluster'] = kmeans
    cluster_groups = test_df2.groupby('cluster')

    # Now print out the details for each cluster
    for _key, group in cluster_groups:
        print('Rows in Cluster: {:d}'.format(len(group)))
        print(group.head(), '\n')

    # Test a corner case (submitted via GitHub)
    cat_df = pd.DataFrame({'a': ['a', 'b', 'c'], 'b': ['v', 'c', 'a']})
    cat_df = cat_df.astype('category')
    copy_cat_df = cat_df.copy()
    print('FIT-TRANSFORM')
    to_matrix = DataFrameToMatrix()
    matrix = to_matrix.fit_transform(cat_df)
    print(matrix)
    print('TRANSFORM')
    matrix_test = to_matrix.transform(cat_df)
    print(matrix_test)

    # These two matrices should be the same
    np_test_utils.assert_equal(matrix, matrix_test)

    # Assert that the dataframe we passed in didn't change
    copy_cat_df.equals(cat_df)


if __name__ == "__main__":
    test()



"""Tor detection and SSL port count example"""

import os
import sys
import argparse
import re
from collections import Counter

# Local imports
from zat import zeek_log_reader

if __name__ == '__main__':
    # Example to check for potential Tor connections and give a summary of different ports
    # used for SSL connections. Please note that your Zeek installation must stamp the
    # ssl.log file with the 'issuer' field. More info can be found here:
    # https://docs.zeek.org/en/master/script-reference/proto-analyzers.html#zeek-ssl

    # Set up the regex search that is used against the issuer field
    issuer_regex = re.compile('CN=www.\w+.com')

    # Set up the regex search that is used against the subject field
    subject_regex = re.compile('CN=www.\w+.net')

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('zeek_log', type=str, help='Specify a zeek log to run ZeekLogReader test on')
    parser.add_argument('-t', action='store_true', default=False, help='Sets the program to tail a live Zeek log')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Sanity check that this is a ssl log
    if 'ssl' not in args.zeek_log:
        print('This example only works with Zeek ssl.log files..')
        sys.exit(1)

    # File may have a tilde in it
    if args.zeek_log:
        args.zeek_log = os.path.expanduser(args.zeek_log)

        # Run the zeek reader on the ssl.log file looking for potential Tor connections
        reader = zeek_log_reader.ZeekLogReader(args.zeek_log, tail=args.t)
        # Just a counter to keep an eye on how many possible Tor connections we identify
        number = 0
        # A empty list to use for the port statistics
        ports = []

        for row in reader.readrows():
            # Add the destination port to the list of ports
            ports.append(row['id.resp_p'])
            # Pull out the Certificate Issuer
            try:
                issuer = row['issuer']
            except KeyError:
                print('Could not find the issuer field in your ssl.log. Please verify your log file.')
                sys.exit(1)
            # Check if the issuer matches the known Tor format
            if issuer_regex.match(issuer):
                # Pull out the Certificate Subject
                try:
                    subject = row['subject']
                except KeyError:
                    print('Could not find the subject field in your ssl.log. Please verify your log file.')
                    sys.exit(1)
                # Check if the subject matches the known Tor format
                if subject_regex.match(subject):
                    print('\nPossible Tor connection found')
                    print('From: {:s} To: {:s} Port: {:d}'.format(row['id.orig_h'], row['id.resp_h'], row['id.resp_p']))
                    number +=1

        # If we are not tailing a live log file, let's print some stats.
        if not args.t:
            # First let's print (if any) the number of possible Tor connections that were found
            print('\nTotal number of possible Tor connections found: {:d}'.format(number))
            # Now let's do the stats on and printing of the port count
            portcount = Counter(ports)
            print('\nPort statistics')
            for port, count in portcount.most_common():
                print('{:<7} {:d}'.format(port, count))



"""Zeek log to CSV file Example"""
import os
import sys
import argparse

# Local imports
from zat.log_to_dataframe import LogToDataFrame


if __name__ == '__main__':
    # Example to write CSV file from a zeek log

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('zeek_log', type=str, help='Specify the zeek log input file')
    parser.add_argument('csv_file', type=str, help='Specify the CSV file to write out')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # File may have a tilde in it
    if args.zeek_log and args.csv_file:
        args.zeek_log = os.path.expanduser(args.zeek_log)
        args.csv_file = os.path.expanduser(args.csv_file)

        # Convert to dataframe and write out the CSV file
        log_to_df = LogToDataFrame()
        zeek_df = log_to_df.create_dataframe(args.zeek_log)
        print('Dataframe Created: {:d} rows...'.format(len(zeek_df)))
        zeek_df.to_csv(args.csv_file, index=False)
        print('Complete: {:s} --> {:s}'.format(args.zeek_log, args.csv_file))



"""JSONLogToDataFrame: Converts a Zeek JSON log to a Pandas DataFrame"""

import os

# Third Party
import pandas as pd

# Local Imports


class JSONLogToDataFrame(object):
    """JSONLogToDataFrame: Converts a Zeek JSON log to a Pandas DataFrame
        Notes:
            Unlike the regular Zeek logs, when you dump the data to JSON you lose
            all the type information. This means we have to guess/infer a lot
            of the types, we HIGHLY recommend that you use the standard Zeek output
            log format as it will result in both faster and better dataframes.
        Todo:
            1. Have a more formal column mapping
            2. Convert Categorial columns
    """
    def __init__(self):
        """Initialize the JSONLogToDataFrame class"""

        # Type conversion Map: This is simple for now but can/should be improved
        self.type_map = {}

    def create_dataframe(self, log_filename, ts_index=True, aggressive_category=True, maxrows=None):
        """ Create a Pandas dataframe from a Zeek JSON log file
            Args:
               log_filename (string): The full path to the Zeek log
               ts_index (bool): Set the index to the 'ts' field (default = True)
               aggressive_category (bool): convert unknown columns to category (default = True)
               maxrows: Read in a subset of rows for testing/inspecting (default = None)
        """
        # Sanity check the filename
        if not os.path.isfile(log_filename):
            print(f'Could not find file: {log_filename}')
            return pd.DataFrame()

        # Read in the JSON file as a dataframe
        _df = pd.read_json(log_filename, nrows=maxrows, lines=True)

        # If we have a ts field convert it to datetime (and optionally set as index)
        if 'ts' in _df.columns:
            _df['ts'] = pd.to_datetime(_df['ts'], unit='s')

            # Set the index
            if ts_index:
                _df.set_index('ts', inplace=True)

        # Okay our dataframe should be ready to go
        return _df


# Simple test of the functionality
def test():
    """Test for JSONLogToDataFrame Class"""
    import os
    pd.set_option('display.width', 1000)
    from zat.utils import file_utils

    # Grab a test file
    data_path = file_utils.relative_dir(__file__, '../data/json')
    log_path = os.path.join(data_path, 'conn.log')

    # Convert it to a Pandas DataFrame
    log_to_df = JSONLogToDataFrame()
    my_df = log_to_df.create_dataframe(log_path)

    # Print out the head
    print(my_df.head())

    # Print out the datatypes
    print(my_df.dtypes)

    # Test a bunch
    tests = ['capture_loss.log', 'dhcp.log', 'http.log', 'ntp.log', 'smb_mapping.log', 'weird.log',
             'conn.log', 'dns.log', 'kerberos.log', 'packet_filter.log', 'ssl.log', 'x509.log',
             'dce_rpc.log', 'files.log', 'loaded_scripts.log', 'smb_files.log', 'stats.log']
    for log_path in [os.path.join(data_path, log) for log in tests]:
        print('Testing: {:s}...'.format(log_path))
        my_df = log_to_df.create_dataframe(log_path)
        print(my_df.head())
        print(my_df.dtypes)

    # Test out maxrows arg
    conn_path = os.path.join(data_path, 'conn.log')
    my_df = log_to_df.create_dataframe(conn_path, maxrows=3)
    print(my_df.head())

    # Test an empty log (a log with header/close but no data rows)
    log_path = os.path.join(data_path, 'http_empty.log')
    my_df = log_to_df.create_dataframe(log_path)

    # Print out the head
    print(my_df.head())

    # Print out the datatypes
    print(my_df.dtypes)

    print('JSONLogToDataFrame Test successful!')


if __name__ == '__main__':
    # Run the test for easy testing/debugging

    # Setup Pandas output options
    pd.options.display.max_colwidth = 20
    pd.options.display.max_columns = 10
    test()



"""Zeek log to Parquet Dataframe Example"""

import os
import sys
import argparse
try:
    from pyspark.sql import SparkSession
except ImportError:
    print('pip install pyspark')
    sys.exit(1)

# Local imports
from zat import log_to_sparkdf


# Helper method
def log_to_parquet(log_in, parquet_out):
    # Spin up a local Spark Session (with 4 executors)
    spark = SparkSession.builder.master('local[4]').appName('my_awesome').getOrCreate()

    # Use the ZAT class to load our log file into a Spark dataframe (2 lines of code!)
    spark_it = log_to_sparkdf.LogToSparkDF(spark)
    spark_df = spark_it.create_dataframe(log_in)

    # Write it out as a parquet file
    spark_df.write.parquet(parquet_out, compression='snappy', mode='overwrite')
    print('{:s} --> {:s}'.format(log_in, parquet_out))

    # Read in the Parquet file
    spark_df = spark.read.parquet(parquet_out)
    spark_df.show(5)


if __name__ == '__main__':
    # Example to write Parquet file from a zeek log

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('zeek_log', type=str, help='Specify the zeek log input file')
    parser.add_argument('parquet_file', type=str, help='Specify the parquet file to write out')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # File may have a tilde in it
    if args.zeek_log and args.parquet_file:
        args.zeek_log = os.path.expanduser(args.zeek_log)
        args.parquet_file = os.path.expanduser(args.parquet_file)

        # Write out the parquet file
        log_to_parquet(args.zeek_log, args.parquet_file)



"""Zeek log to Pandas Dataframe + Filter by WhiteList Example"""

import os
import sys
import argparse
import pandas as pd

# Local imports
from zat.log_to_dataframe import LogToDataFrame

if __name__ == '__main__':
    # Example to populate a Pandas dataframe from a zeek log reader

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('dns_log', help='Specify the zeek DNS log')
    parser.add_argument('whitelist', help='Specify the DNS whiteliist')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Grab the whitelist
    white_df = pd.read_csv(args.whitelist, names=['rank', 'domain'])
    whitelist = white_df['domain'].tolist()

    # File may have a tilde in it
    if args.dns_log:
        args.dns_log = os.path.expanduser(args.dns_log)

        # Create a Pandas dataframe from a Zeek log
        log_to_df = LogToDataFrame()
        zeek_df = log_to_df.create_dataframe(args.dns_log)

        # Print out the head of the dataframe
        print('DF Size before whitelist: {:d} rows'.format(len(zeek_df)))

        # Filter the dataframe with the whitelist
        white_df = zeek_df[zeek_df['query'].isin(whitelist)]
        print('Filtering out {!r}'.format(white_df['query'].tolist()))
        zeek_df = zeek_df[~zeek_df['query'].isin(whitelist)]

        print('DF Size after whitelist: {:d} rows'.format(len(zeek_df)))



"""Risky DNS zat Example"""

import sys
import argparse
from pprint import pprint
import pickle
import json
from kafka import KafkaConsumer
from kafka.errors import NoBrokersAvailable

# Third Party Imports
try:
    import tldextract
except ImportError:
    print('\nThis example needs tldextract. Please do a $pip install tldextract and rerun this example')
    sys.exit(1)

# Local imports
from zat.utils import vt_query, signal_utils


def save_vtq():
    """Exit on Signal"""
    global vtq

    print('Saving VirusTotal Query Cache...')
    pickle.dump(vtq, open('vtq.pkl', 'wb'), protocol=pickle.HIGHEST_PROTOCOL)
    sys.exit()


if __name__ == '__main__':
    # Risky DNS/VT Query application
    global vtq

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('--server', type=str, default='localhost:9092',
                        help='Specify the Kafka Server (default: localhost:9092)')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # First we create a Kafka Consumer
    kserver = args.server
    try:
        consumer = KafkaConsumer('dns', bootstrap_servers=[kserver],
                                 value_deserializer=lambda x: json.loads(x.decode('utf-8')))
    except NoBrokersAvailable:
        print('Could not connect to Kafka server: {:s}'.format(args.server))
        sys.exit(-1)

    # See if we have a serialized VirusTotal Query Class.
    # If we do not have one we'll create a new one
    try:
        vtq = pickle.load(open('vtq.pkl', 'rb'))
        print('Opening VirusTotal Query Cache (cache_size={:d})...'.format(vtq.size))
    except IOError:
        vtq = vt_query.VTQuery(max_cache_time=60*24*7)  # One week cache

    # See our 'Risky Domains' Notebook for the analysis and
    # statistical methods used to compute this risky set of TLDs
    risky_tlds = set(['info', 'tk', 'xyz', 'online', 'club', 'ru', 'website', 'in', 'ws',
                      'top', 'site', 'work', 'biz', 'name', 'tech', 'loan', 'win', 'pro'])

    # Launch long lived process with signal catcher
    with signal_utils.signal_catcher(save_vtq):

        # Now lets process our Kafka 'dns' Messages
        for message in consumer:
            dns_message = message.value

            # Pull out the TLD
            query = dns_message.get('query')
            tld = tldextract.extract(query).suffix if query else None

            # Check if the TLD is in the risky group
            if tld in risky_tlds:
                print('\n'+query)
                # Make the query with the full query
                results = vtq.query_url(query)
                if results.get('positives', 0) > 3:  # At least four hits
                    print('Risky Domain DNS Query Found')
                    print('From: {:s} To: {:s} QType: {:s} RCode: {:s}'.format(dns_message['id.orig_h'],
                                                                               dns_message['id.resp_h'],
                                                                               dns_message['qtype_name'],
                                                                               dns_message['rcode_name']))
                    pprint(results)



"""Read Kafka Streams into Spark, perform simple filtering/aggregation"""

import sys
import argparse
from time import sleep

try:
    from pyspark.sql import SparkSession
    from pyspark.sql.types import StructType, StringType, BooleanType, IntegerType
    from pyspark.sql.functions import from_json, to_json, col, struct, udf
except ImportError:
    print('\npip install pyspark')
    sys.exit(1)

try:
    import tldextract
except ImportError:
    print('\nThis example needs tldextract. Please do a $pip install tldextract and rerun this example')
    sys.exit(1)


def exit_program():
    """Exit on Signal"""
    print('Exiting Program...')
    sys.exit()


def compute_domain(query):
    # Pull out the domain
    if query.endswith('.local'):
        return 'local'
    return tldextract.extract(query).registered_domain if query else None


if __name__ == '__main__':
    """Read Kafka Streams into Spark, perform simple filtering/aggregation"""
    parser = argparse.ArgumentParser()
    parser.add_argument('--server', type=str, default='localhost:9092',
                        help='Specify the Kafka Server (default: localhost:9092)')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Grab the Kafka server
    kserver = args.server

    # Spin up a local Spark Session (with 4 executors)
    spark = SparkSession.builder.master('local[4]').appName('my_awesome') \
            .config('spark.jars.packages', 'org.apache.spark:spark-sql-kafka-0-10_2.11:2.4.4') \
            .getOrCreate()
    spark.sparkContext.setLogLevel('ERROR')

    # Optimize the conversion to Spark
    spark.conf.set("spark.sql.execution.arrow.enable", "true")

    # SUBSCRIBE: Setup connection to Kafka Stream 
    raw_data = spark.readStream.format('kafka').option('kafka.bootstrap.servers', kserver) \
                                               .option('subscribe', 'dns') \
                                               .option('startingOffsets', 'earliest').load()

    # Define the schema for the DNS message (do this better)
    dns_schema = StructType().add('ts', StringType()).add('uid', StringType()).add('id.orig_h', StringType()) \
                             .add('id.orig_p', IntegerType()).add('id.resp_h', StringType()).add('id.resp_p', IntegerType()) \
                             .add('proto', StringType()).add('trans_id', IntegerType()).add('query', StringType()) \
                             .add('qclass', IntegerType()).add('qclass_name', StringType()).add('qtype', IntegerType()) \
                             .add('qtype_name', StringType()).add('rcode', IntegerType()).add('rcode_name', StringType()) \
                             .add('AA', BooleanType()).add('TC', BooleanType()).add('RD', BooleanType()).add('RA', BooleanType()) \
                             .add('Z', IntegerType()).add('answers', StringType()).add('TTLs', StringType()).add('rejected', BooleanType())

    # ETL: Convert raw data into parsed and proper typed data
    parsed_data = raw_data.select(from_json(col('value').cast('string'), dns_schema).alias('data')).select('data.*')

    # FILTER: Only get DNS records that have 'query' field filled out
    filtered_data = parsed_data.filter(parsed_data.query.isNotNull() & (parsed_data.query!='')==True)

    # FILTER 2: Remove Local/mDNS queries
    filtered_data = filtered_data.filter(~filtered_data.query.like('%.local'))  # Note: using the '~' negation operator

    # COMPUTE: A new column with the 2nd level domain extracted from the query
    udf_compute_domain = udf(compute_domain, StringType())
    computed_data = filtered_data.withColumn('domain', udf_compute_domain('query'))

    # AGGREGATE: In this case a simple groupby operation
    group_data = computed_data.groupBy('`id.orig_h`', 'domain', 'qtype_name').count()

    # At any point in the pipeline you can see what you're getting out
    group_data.printSchema()

    # Take the end of our pipeline and pull it into memory
    dns_count_memory_table = group_data.writeStream.format('memory').queryName('dns_counts').outputMode('complete').start()

    # Let the pipeline pull some data
    print('Pulling pipline...Please wait...')

    # Create a Pandas Dataframe by querying the in memory table and converting
    # Loop around every 5 seconds to update output
    for _ in range(10):
        sleep(5)
        dns_counts_df = spark.sql("select * from dns_counts").toPandas()
        print('\nDNS Query Total Counts = {:d}'.format(dns_counts_df['count'].sum()))
        print(dns_counts_df.sort_values(ascending=False, by='count'))

    # Stop the stream
    dns_count_memory_table.stop()
    sleep(1)


"""FileTailer Python Class"""

import os
import sys
import argparse
import time
import math
from collections import Counter

# Third Party Imports
import pandas as pd
from sklearn.ensemble import IsolationForest
from sklearn.cluster import MiniBatchKMeans

# Local imports
from zat import zeek_log_reader, live_simulator
from zat import dataframe_to_matrix, dataframe_cache


def entropy(string):
    """Compute entropy on the string"""
    p, lns = Counter(string), float(len(string))
    return -sum(count/lns * math.log(count/lns, 2) for count in p.values())


if __name__ == '__main__':
    # Example to show the dataframe cache functionality on streaming data
    pd.set_option('display.width', 200)

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('zeek_log', type=str, help='Specify a zeek log to run ZeekLogReader test on')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # File may have a tilde in it
    if args.zeek_log:
        args.zeek_log = os.path.expanduser(args.zeek_log)

        # Sanity check dns log
        if 'dns' in args.zeek_log:
            log_type = 'dns'
        else:
            print('This example only works with Zeek with dns.log files..')
            sys.exit(1)

        # Create a Zeek log reader
        print('Opening Data File: {:s}'.format(args.zeek_log))
        reader = zeek_log_reader.ZeekLogReader(args.zeek_log, tail=True)

        # Create a Zeek IDS log live simulator
        print('Opening Data File: {:s}'.format(args.zeek_log))
        reader = live_simulator.LiveSimulator(args.zeek_log, eps=10)  # 10 events per second

        # Create a Dataframe Cache
        df_cache = dataframe_cache.DataFrameCache(max_cache_time=600)  # 10 minute cache

        # Streaming Clustering Class
        batch_kmeans = MiniBatchKMeans(n_clusters=5, verbose=True)

        # Use the ZeekThon DataframeToMatrix class
        to_matrix = dataframe_to_matrix.DataFrameToMatrix()

        # Add each new row into the cache
        time_delta = 10
        timer = time.time() + time_delta
        FIRST_TIME = True
        for row in reader.readrows():
            df_cache.add_row(row)

            # Every 30 seconds grab the dataframe from the cache
            if time.time() > timer:
                timer = time.time() + time_delta

                # Get the windowed dataframe (10 minute window)
                zeek_df = df_cache.dataframe()

                # Compute some addition data
                zeek_df['query_length'] = zeek_df['query'].str.len()
                zeek_df['answer_length'] = zeek_df['answers'].str.len()
                zeek_df['entropy'] = zeek_df['query'].map(lambda x: entropy(x))

                # Use the zat DataframeToMatrix class
                features = ['Z', 'proto', 'qtype_name', 'query_length', 'answer_length', 'entropy', 'id.resp_p']
                to_matrix = dataframe_to_matrix.DataFrameToMatrix()
                zeek_matrix = to_matrix.fit_transform(zeek_df[features])
                print(zeek_matrix.shape)

                # Print out the range of the daterange and some stats
                print('DataFrame TimeRange: {:s} --> {:s}'.format(str(zeek_df['ts'].min()), str(zeek_df['ts'].max())))

                # Train/fit and Predict anomalous instances using the Isolation Forest model
                odd_clf = IsolationForest(contamination=0.2)  # Marking 20% as odd
                predictions = odd_clf.fit_predict(zeek_matrix)
                odd_df = zeek_df[predictions == -1]

                # Now we're going to explore our odd observations with help from KMeans
                odd_matrix = to_matrix.transform(odd_df[features])
                batch_kmeans.partial_fit(odd_matrix)
                clusters = batch_kmeans.predict(odd_matrix).tolist()
                odd_df['cluster'] = clusters

                # Now group the dataframe by cluster
                cluster_groups = odd_df.groupby('cluster')

                # Now print out the details for each cluster
                show_fields = ['id.orig_h', 'id.resp_h', 'query'] + features
                print('<<< Outliers Detected! >>>')
                for key, group in cluster_groups:
                    print('\nCluster {:d}: {:d} observations'.format(key, len(group)))
                    print(group[show_fields].head())



"""NGram utilities that might be useful"""


def compute_ngrams(word_list, S=3, T=3):
    """Compute NGrams in the word_list from [S-T)
        Args:
            word_list (list): A list of words to compute ngram set from
            S (int): The smallest NGram (default=3)
            T (int): The biggest NGram (default=3)
    """
    _ngrams = []
    if isinstance(word_list, str):
        word_list = [word_list]
    for word in word_list:
        for n in range(S, T+1):
            _ngrams += zip(*(word[i:] for i in range(n)))
    return [''.join(_ngram) for _ngram in _ngrams]


def ngram_count(word, ngrams):
    """Compute the number of matching NGrams in the given word"""
    return len(set(ngrams).intersection(compute_ngrams([word])))


def test():
    """Test the ngram methods"""
    domains = ['google', 'facebook', 'apple']
    compute_ngrams(domains, 2, 5)
    ngrams = compute_ngrams(domains)
    print('NGrams: {:s}'.format(str(list(ngrams))))
    print(ngram_count('foogle', ngrams))
    print(ngram_count('mybook', ngrams))


if __name__ == '__main__':
    test()



"""Zeek log to Pandas Dataframe Example"""

import os
import sys
import argparse

# Local imports
from zat.log_to_dataframe import LogToDataFrame

if __name__ == '__main__':
    # Example to populate a Pandas dataframe from a zeek log reader

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('zeek_log', type=str, help='Specify a zeek log to run ZeekLogReader test on')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # File may have a tilde in it
    if args.zeek_log:
        args.zeek_log = os.path.expanduser(args.zeek_log)

        # Create a Pandas dataframe from a Zeek log
        log_to_df = LogToDataFrame()
        zeek_df = log_to_df.create_dataframe(args.zeek_log)

        # Print out the head of the dataframe
        print(zeek_df.head())

        # Print out the types of the columns
        print(zeek_df.dtypes)

        # Print out size and memory usage
        print('DF Shape: {:s}'.format(str(zeek_df.shape)))
        print('DF Memory:')
        memory_usage = zeek_df.memory_usage(deep=True)
        total = memory_usage.sum()
        for item in memory_usage.items():
            print('\t {:s}: \t{:.2f} MB'.format(item[0], item[1]/1e6))
        print('DF Total: {:.2f} GB'.format(total/(1e9)))



# Example that demonstrates going from Zeek data to scikit-learn models

import os
import sys
import argparse

# Third Party Imports
import pandas as pd
from sklearn.decomposition import PCA
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.cluster import KMeans
import numpy as np

# Local imports
from zat import log_to_dataframe
from zat import dataframe_to_matrix

# Helper method for scatter/beeswarm plot
def jitter(arr):
    stdev = .02*(max(arr)-min(arr))
    return arr + np.random.randn(len(arr)) * stdev

if __name__ == '__main__':
    # Example that demonstrates going from Zeek data to scikit-learn models

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('zeek_log', type=str, help='Specify a zeek log to run ZeekLogReader test on')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Sanity check that this is a dns log
    if 'dns' not in args.zeek_log:
        print('This example only works with Zeek dns.log files..')
        sys.exit(1)

    # File may have a tilde in it
    if args.zeek_log:
        args.zeek_log = os.path.expanduser(args.zeek_log)

        # Create a Pandas dataframe from the Zeek log
        log_to_df = log_to_dataframe.LogToDataFrame()
        zeek_df = log_to_df.create_dataframe(args.zeek_log)

        # Add query length
        zeek_df['query_length'] = zeek_df['query'].str.len()

        # Normalize this field
        #ql = zeek_df['query_length']
        #zeek_df['query_length_norm'] = (ql - ql.min()) / (ql.max()-ql.min())

        # These are the features we want (note some of these are categorical!)
        features = ['AA', 'RA', 'RD', 'TC', 'Z', 'rejected', 'proto', 'qtype_name', 'rcode_name', 'query_length']
        feature_df = zeek_df[features]

        # Use the super awesome DataframeToMatrix class (handles categorical data!)
        to_matrix = dataframe_to_matrix.DataFrameToMatrix()
        zeek_matrix = to_matrix.fit_transform(feature_df)

        # Now we're ready for scikit-learn!
        # Just some simple stuff for this example, KMeans and PCA
        kmeans = KMeans(n_clusters=5).fit_predict(zeek_matrix)
        pca = PCA(n_components=2).fit_transform(zeek_matrix)

        # Now we can put our ML results back onto our dataframe!
        zeek_df['x'] = jitter(pca[:, 0]) # PCA X Column
        zeek_df['y'] = jitter(pca[:, 1]) # PCA Y Column
        zeek_df['cluster'] = kmeans

        # Now use dataframe group by cluster
        show_fields = ['query', 'Z', 'proto', 'qtype_name', 'x', 'y', 'cluster']
        cluster_groups = zeek_df[show_fields].groupby('cluster')

        # Now print out the details for each cluster
        pd.set_option('display.width', 1000)
        for key, group in cluster_groups:
            print('Rows in Cluster: {:d}'.format(len(group)))
            print(group.head(), '\n')



""" DataFrame Statistics Methods
        - Contingency Table (also called Cross Tabulation)
        - Joint Distribution
        - G-Scores
    References:
        - http://en.wikipedia.org/wiki/Contingency_table
        - http://en.wikipedia.org/wiki/G_test (Wikipedia)
        - http://udel.edu/~mcdonald/stathyptesting.html (Hypothesis Testing)
"""

import math

# Third Party
import pandas as pd


def contingency_table(dataframe, rownames, colnames, margins=True):
    """Contingency Table (also called Cross Tabulation)
       - Table in a matrix format that displays the (multivariate) frequency distribution of the variables
       - http://en.wikipedia.org/wiki/Contingency_table
       Args:
           rownames: the column name or list of columns names that make the keys of the rows
           colnames: the column name or list of columns names that make the keys of the columns
    """
    # Taking just the rownames + colnames of the dataframe
    sub_set = [rownames, colnames]
    _sub_df = dataframe[sub_set]
    return _sub_df.pivot_table(index=rownames, columns=colnames, margins=margins, aggfunc=len, fill_value=0)


def joint_distribution(dataframe, rownames, colnames):
    """Joint Distribution Table
       - The Continguency Table normalized by the total number of observations
       Args:
           rownames: the column name or list of columns names that make the keys of the rows
           colnames: the column name or list of columns names that make the keys of the columns
    """
    cont_table = contingency_table(dataframe, rownames=rownames, colnames=colnames, margins=True)
    total_observations = cont_table['All']['All']
    return cont_table/total_observations


def expected_counts(dataframe, rownames, colnames):
    """Expected counts of the multivariate frequency distribution of the variables given the
       null hypothesis of complete independence between variables.
       Args:
           rownames: the column name or list of columns names that make the keys of the rows
           colnames: the column name or list of columns names that make the keys of the columns
    """
    cont_table = contingency_table(dataframe, rownames=rownames, colnames=colnames, margins=True)
    row_counts = cont_table['All']
    column_counts = cont_table.loc['All']
    total_observations = cont_table['All']['All']

    # There didn't seem to be a good way to vectorize this (Fixme?)
    for column in cont_table.columns:
        for row in cont_table.index:
            cont_table[column][row] = column_counts[column]*row_counts[row]/total_observations
    return cont_table


def g_test_scores(dataframe, rownames, colnames):
    """G Test Score for log likelihood ratio
       - http://en.wikipedia.org/wiki/G_test (Wikipedia)
       - 95th percentile; 5% level; p < 0.05; critical value = 3.84
       - 99th percentile; 1% level; p < 0.01; critical value = 6.63
       - 99.9th percentile; 0.1% level; p < 0.001; critical value = 10.83
       - 99.99th percentile; 0.01% level; p < 0.0001; critical value = 15.13

       Args:
           rownames: the column name or list of columns names that make the keys of the rows
           colnames: the column name or list of columns names that make the keys of the columns
    """
    cont_table = contingency_table(dataframe, rownames=rownames, colnames=colnames, margins=False)
    exp_counts = expected_counts(dataframe, rownames=rownames, colnames=colnames)

    # There didn't seem to be a good way to vectorize this (Fixme?)
    for row in cont_table.index:
        g_score = 0
        for column in cont_table.columns:
            g_score += compute_g(cont_table[column][row], exp_counts[column][row])
        for column in cont_table.columns:
            cont_table[column][row] = g_score if cont_table[column][row] > exp_counts[column][row] else -g_score
    return cont_table


def compute_g(count, expected):
    """G Test Score for log likelihood ratio
       - http://en.wikipedia.org/wiki/G_test (Wikipedia)
    """
    try:
        return 2.0 * count * math.log(count/expected)
    except ValueError:
        return 0


# Simple test of the functionality
def test():
    """Test for DataFrame Stats module"""

    import os
    from zat.utils import file_utils

    # Open a dataset (relative path)
    data_dir = file_utils.relative_dir(__file__, 'test_data')
    file_path = os.path.join(data_dir, 'g_test_data.csv')
    dataframe = pd.read_csv(file_path)
    print(dataframe.head())

    # Print out the contingency_table
    print('\nContingency Table')
    print(contingency_table(dataframe, 'name', 'status'))

    # Print out the joint_distribution
    print('\nJoint Distribution Table')
    print(joint_distribution(dataframe, 'name', 'status'))

    # Print out the expected_counts
    print('\nExpected Counts Table')
    print(expected_counts(dataframe, 'name', 'status'))

    # Print out the g_test scores
    print('\nG-Test Scores')
    print(g_test_scores(dataframe, 'name', 'status'))


if __name__ == "__main__":
    test()



"""LogToSparkDF: Converts a Zeek log to a Spark DataFrame"""

import sys

# Third Party
try:
    from pyspark.sql.types import StructType, StringType, IntegerType, FloatType, LongType
    from pyspark.sql.functions import col, when
except ImportError:
    print('\npip install pyspark')


# Local
from zat import zeek_log_reader


class LogToSparkDF(object):
    """LogToSparkDF: Converts a Zeek log to a Spark DataFrame"""

    def __init__(self, spark):
        """Initialize the LogToSparkDF class"""

        # Grab the spark context
        self.spark = spark

        # First Level Type Mapping
        #    This map defines the types used when first reading in the Zeek log into a 'chunk' dataframes.
        #    Types (like time and interval) will be defined as one type at first but then
        #    will undergo further processing to produce correct types with correct values.
        # See: https://spark.apache.org/docs/latest/sql-reference.html
        #      for more info on supported types.
        self.type_map = {'bool': StringType(),   # Secondary Processing into BooleanType()
                         'count': LongType(),
                         'int': IntegerType(),
                         'double': FloatType(),
                         'time': FloatType(),    # Secondary Processing into TimestampType()
                         'interval': FloatType(),
                         'port': IntegerType(),
                         'enum': StringType(),
                         'addr': StringType(),
                         'string': StringType()
                         }

    def create_dataframe(self, log_filename, fillna=True):
        """ Create a Spark dataframe from a Bro/Zeek log file
            Args:
               log_fllename (string): The full path to the Zeek log
               fillna (bool): Fill in NA/NaN values (default=True)
        """

        # Create a Zeek log reader just to read in the header for names and types
        _zeek_reader = zeek_log_reader.ZeekLogReader(log_filename)
        _, field_names, field_types, _ = _zeek_reader._parse_zeek_header(log_filename)

        # Get the appropriate types for the Spark Dataframe
        spark_schema = self.build_spark_schema(field_names, field_types)

        # Now actually read the Zeek Log using Spark read CSV
        _df = self.spark.read.csv(log_filename, schema=spark_schema, sep='\t', comment="#", nullValue='-')

        ''' Secondary processing (cleanup)
            - Fix column names with '.' in them
            - Fill in Nulls (optional)
            - timestamp convert
            - boolean convert
        '''

        # Fix column names
        ''' Note: Yes column names with '.' in them can be escaped with backticks when selecting them BUT
                  many pipeline operations will FAIL internally if the column names have a '.' in them.
        '''
        fixed_columns = list(map(lambda x: x.replace('.', '_'), _df.columns))
        _df = _df.toDF(*fixed_columns)

        # Fill in NULL values
        if fillna:
            _df = _df.na.fill(0)    # For numeric columns
            _df = _df.na.fill('-')  # For string columns

        # Convert timestamp and boolean columns
        for name, f_type in zip(field_names, field_types):
            # Some field names may have '.' in them, so we create a reference name to those fields
            ref_name = name.replace('.', '_')
            if f_type == 'time':
                _df = _df.withColumn(name, _df[ref_name].cast('timestamp'))
            if f_type == 'bool':
                _df = _df.withColumn(name, when(col(ref_name) == 'T', 'true').when(col(ref_name) == 'F', 'false')
                                     .otherwise('null').cast('boolean'))

        # Return the spark dataframe
        return _df

    def build_spark_schema(self, column_names, column_types, verbose=False):
        """Given a set of names and types, construct a dictionary to be used
           as the Spark read_csv dtypes argument"""

        # If we don't know the type put it into a string
        unknown_type = StringType()

        schema = StructType()
        for name, zeek_type in zip(column_names, column_types):

            # Grab the type
            spark_type = self.type_map.get(zeek_type)

            # Sanity Check
            if not spark_type:
                if verbose:
                    print('Could not find type for {:s} using StringType...'.format(zeek_type))
                spark_type = unknown_type

            # Add the Spark type for this column
            schema.add(name, spark_type)

        # Return the Spark schema
        return schema


# Simple test of the functionality
def test():
    """Test for LogToSparkDF Class"""
    import os
    import pytest
    from zat.utils import file_utils

    try:
        from pyspark.sql import SparkSession
    except ImportError:
        pytest.skip('pip install pyspark')

    # Spin up a local Spark Session (with 4 executors)
    spark = SparkSession.builder.master('local[4]').appName('my_awesome').getOrCreate()

    # Grab a test file
    data_path = file_utils.relative_dir(__file__, '../data')
    log_path = os.path.join(data_path, 'ftp.log')

    # Convert it to a Spark DataFrame
    log_to_spark = LogToSparkDF(spark)
    spark_df = log_to_spark.create_dataframe(log_path)

    # Print out the head
    print(spark_df.show())

    # Print out the datatypes
    print(spark_df.printSchema())

    num_rows = spark_df.count()
    print("Number of Spark DataFrame rows: {:d}".format(num_rows))
    columns = spark_df.columns
    print("Columns: {:s}".format(','.join(columns)))

    # Test a bunch
    tests = ['app_stats.log', 'dns.log', 'http.log', 'notice.log', 'tor_ssl.log',
             'conn.log', 'dhcp.log', 'dhcp_002.log', 'files.log',  'smtp.log', 'weird.log',
             'ftp.log',  'ssl.log', 'x509.log']
    for log_path in [os.path.join(data_path, log) for log in tests]:
        print('Testing: {:s}...'.format(log_path))
        spark_df = log_to_spark.create_dataframe(log_path)
        print(spark_df.show())
        print(spark_df.printSchema())

    # Test an empty log (a log with header/close but no data rows)
    log_path = os.path.join(data_path, 'http_empty.log')
    spark_df = log_to_spark.create_dataframe(log_path)
    print(spark_df.show())
    print(spark_df.printSchema())

    print('LogToSparkDF Test successful!')


if __name__ == '__main__':
    # Run the test for easy testing/debugging
    test()



"""FileStorage class for storing/retrieving bytes to/from persistant storage
      Methods:
        - store(key, bytes): Takes a bytes object as input
        - get(key): Returns a bytes object
"""

import os

try:
    import pyarrow  # noqa F401
    _HAVE_PYARROW = True
except ImportError:
    print('pyarrow not found, $ pip install pyarrow for more tests...')
    _HAVE_PYARROW = False


class FileStorage(object):
    """FileStorage class for storing/retrieving bytes to/from persistant storage
          Methods:
            - store(key, bytes): Takes a bytes object as input
            - get(key): Returns a bytes object
     """

    def __init__(self):
        """FileStorage Initialization"""
        self.tmp_dir = '/tmp/zat_file_storage'
        os.makedirs(self.tmp_dir, exist_ok=True)

    def store(self, key, bytes_buffer):
        """Store the buffer with the associated key"""

        # Write the temporary file
        try:
            filename = self.compute_filename(key)
            tempfile = filename + '.tmp'
            with open(tempfile, 'wb') as fp:
                fp.write(bytes_buffer)
            os.rename(tempfile, filename)
        except (PermissionError, IOError):
            msg = 'FileStorage: could not write to disk!'
            print(msg)
            raise IOError(msg)

    def get(self, key):
        """Retrieve the buffer associated with the given key"""

        # Now see if we can read it off disk (it may have been removed/expired)
        try:
            filename = self.compute_filename(key)
            print('FileStorage: Returning bytes for: {:s}'.format(key))
            with open(filename, 'rb') as fp:
                return fp.read()

        except IOError:
            print('Could not read file for key: {:s}'.format(key))
            return None

    def compute_filename(self, key):
        # Compute the temporary file name
        return os.path.join(self.tmp_dir, key)

    def stored_files(self):
        return [os.path.join(self.tmp_dir, name) for name in os.listdir(self.tmp_dir)]

    @property
    def size(self):
        """Return size of the query/dataframe store"""
        return len(self.stored_files())

    def clear(self):
        """Clear the query/dataframe store"""
        for filename in self.stored_files():
            try:
                os.unlink(filename)
            except IOError:
                print('Could not delete: {:s}'.format(filename))

    def dump(self):
        """Dump the cache key/values (for debugging)"""
        for filename in self.stored_files():
            print(filename)


def test():
    """Test for the FileStorage class"""
    import json
    from io import BytesIO
    import pandas as pd

    # Create some data
    data1 = {'foo': [1, 2, 1, 1, 2, 3], 'name': ['bob', 'bob', 'sue', 'sue', 'jim', 'jim']}
    data2 = {'count': [8, 9, 8, 8, 8, 9], 'name': ['joe', 'sal', 'joe', 'sal', 'joe', 'sal']}
    my_storage = FileStorage()
    my_storage.clear()

    # Serialize the data
    bytes1 = json.dumps(data1).encode('utf-8')
    bytes2 = json.dumps(data2).encode('utf-8')

    # Test storage
    my_storage.store('data1_key', bytes1)
    my_storage.store('data2_key', bytes2)

    # Make sure size is working
    assert my_storage.size == 2

    # Try grabbing a key that doesn't exist
    assert my_storage.get('no key') is None

    # Dump the storage
    my_storage.dump()

    # Retrieve the stored data
    r_data1 = json.loads(my_storage.get('data1_key'))
    r_data2 = json.loads(my_storage.get('data2_key'))
    assert r_data1 == data1
    assert r_data2 == data2

    # Delete our key value entries
    my_storage.clear()
    assert my_storage.size == 0

    # Dump the cache
    my_storage.dump()

    # Now run all the same tests with dataframes (only if pyarrow available)
    if _HAVE_PYARROW:

        # Helper methods
        def dataframe_to_bytes(df):
            bytes_buffer = BytesIO()
            df.to_parquet(bytes_buffer)
            return bytes_buffer.getvalue()

        def dataframe_from_bytes(df_bytes):
            return pd.read_parquet(BytesIO(df_bytes))

        # Create the a dataframe and a DataStore class
        df1 = pd.DataFrame(data={'foo': [1, 2, 1, 1, 2, 3], 'name': ['bob', 'bob', 'sue', 'sue', 'jim', 'jim']})
        df2 = pd.DataFrame(data={'count': [8, 9, 8, 8, 8, 9], 'name': ['joe', 'sal', 'joe', 'sal', 'joe', 'sal']})
        my_storage.clear()

        # Serialize the dataframes
        df1_bytes = dataframe_to_bytes(df1)
        df2_bytes = dataframe_to_bytes(df2)

        # Test storage
        my_storage.store('df1_key', df1_bytes)
        my_storage.store('df2_key', df2_bytes)

        # Make sure size is working
        assert my_storage.size == 2

        # Dump the cache
        my_storage.dump()

        # Retrieve the cached dataframes
        r_df1 = dataframe_from_bytes(my_storage.get('df1_key'))
        r_df2 = dataframe_from_bytes(my_storage.get('df2_key'))
        assert r_df1.equals(df1)
        assert r_df2.equals(df2)

        # Delete our key value entries
        my_storage.clear()
        assert my_storage.size == 0


if __name__ == '__main__':

    # Run the test
    test()



"""FileTailer Python Class"""

import os
import time

# Local imports
from zat.utils import file_utils


class FileTailer(object):
    """FileTailer: Will provide 'tail -f' functionality for a file. The readlines() method
                   returns a generator that yields lines as they are added to the file

        Args:
            filepath (str): The full path the file (/full/path/to/the/file.txt)
            sleep (int): The wait interval in milliseconds (default=50)
            full_read (bool): Read the full file  (default=True)
            tail (bool): Do a dynamic tail on the file (i.e. tail -f) (default=True)
    """
    def __init__(self, filepath, sleep=50, full_read=True, tail=True):
        """FileTailer Initialization"""
        self._filepath = filepath
        self._sleep = sleep * 1e-3
        self._full_read = full_read
        self._tail = tail

    def readlines(self, offset=0):
        """Open the file for reading and yield lines as they are added"""
        try:
            with open(self._filepath) as fp:
                # For full read go through existing lines in file
                if self._full_read:
                    fp.seek(offset)
                    for row in fp:
                        yield row

                # Okay now dynamically tail the file
                if self._tail:
                    while True:
                        current = fp.tell()
                        row = fp.readline()
                        if row:
                            yield row
                        else:
                            fp.seek(current)
                            time.sleep(self._sleep)

        except IOError as err:
            print('Error reading the file {0}: {1}'.format(self._filepath, err))
            return


def test():
    """Test for FileTailer Python Class"""

    # Grab a test file
    data_path = file_utils.relative_dir(__file__, '../../data')
    test_path = os.path.join(data_path, 'http.log')
    print('Opening Data File: {:s}'.format(test_path))

    # Create the Class
    tailer = FileTailer(test_path, tail=False)  # First with no tailing
    for line in tailer.readlines():
        print(line)
    print('Read with NoTail Test successful!')

    # Now include tailing (note: as an automated test this needs to timeout quickly)
    try:
        from interruptingcow import timeout

        # Spin up the class
        tailer = FileTailer(test_path)  # Tail = True

        # Tail the file for 2 seconds and then quit
        try:
            with timeout(2, exception=RuntimeError):
                for line in tailer.readlines():
                    print(line)
        except RuntimeError:  # InterruptingCow raises a RuntimeError on timeout
            print('Tailing Test successful!')

    except ImportError:
        print('Tailing Test not run, need interruptcow module...')


if __name__ == '__main__':
    # Run the test for easy testing/debugging
    test()



#!/usr/bin/env python
"""Setup.py for zat"""

import os

from setuptools import setup, find_packages

readme = open('README.md').read()

# Pull in the package info
package_name = 'zat'
package = __import__(package_name)
version = package.__version__
author = package.__author__
email = package.__email__


# Data and Example Files
def get_files(dir_name):
    """Simple directory walker"""
    return [(os.path.join('.', d), [os.path.join(d, f) for f in files]) for d, _, files in os.walk(dir_name)]


setup(
    name=package_name,
    version=version,
    description='Zeek Analysis Tools',
    long_description=readme,
    long_description_content_type='text/markdown',
    author=author,
    author_email=email,
    url='https://github.com/SuperCowPowers/zat',
    packages=find_packages(),
    include_package_data=True,
    data_files=get_files('data') + get_files('examples'),
    install_requires=[
        'requests',
        'watchdog',
        'pandas',
        'scikit-learn'
    ],
    extras_require={
        'pyspark':  ['pyspark'],
        'all':  ['pyspark', 'pyarrow', 'yara-python', 'tldextract']
    },
    license='Apache',
    keywords='Zeek, Bro, Python, Networking, Security, Scikit-Learn, Spark, Kafka, Parquet',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
)



"""ReverseDNS, Perform a reverse dns lookup on fields in the ip_field_list."""

import socket

# Local imports
from zat.utils import net_utils, cache


class ReverseDNS(object):
    """Perform a reverse dns lookup on fields in the ip_field_list."""

    def __init__(self, lookup_internal=False):
        """Initialize ReverseDNS Class"""
        self.ip_lookup_cache = cache.Cache(timeout=600)
        self.lookup_internal = lookup_internal

    def lookup(self, ip_address):
        """Try to do a reverse dns lookup on the given ip_address"""

        # Is this already in our cache
        if self.ip_lookup_cache.get(ip_address):
            domain = self.ip_lookup_cache.get(ip_address)

        # Is the ip_address local or special
        elif not self.lookup_internal and net_utils.is_internal(ip_address):
            domain = 'internal'
        elif net_utils.is_special(ip_address):
            domain = net_utils.is_special(ip_address)

        # Look it up at this point
        else:
            domain = self._reverse_dns_lookup(ip_address)

        # Cache it
        self.ip_lookup_cache.set(ip_address, domain)

        # Return the domain
        return domain

    @staticmethod
    def _reverse_dns_lookup(ip_address):
        """Perform the reverse DNS lookup

           Args:
               ip_address: the ip_address (as a str)
           Returns:
               the domain given by a reverse DNS request on the ip address
        """

        # Look it up
        try:
            return socket.gethostbyaddr(ip_address)[0]
        except socket.herror:
            return 'nxdomain'


def test():
    """Test for ReverseDNS class"""
    dns = ReverseDNS()

    print(dns.lookup('192.168.0.1'))
    print(dns.lookup('8.8.8.8'))

    # Test cache
    print(dns.lookup('8.8.8.8'))


if __name__ == '__main__':
    test()



"""Network utilities that might be useful"""


import socket
import binascii
import ipaddress


def mac_to_str(address):
    """Convert a MAC address to a readable/printable string

       Args:
           address (str): a MAC address in hex form (e.g. '\x01\x02\x03\x04\x05\x06')
       Returns:
           str: Printable/readable MAC address
    """
    return ':'.join('%02x' % b for b in address)


def str_to_mac(mac_string):
    """Convert a readable string to a MAC address

           Args:
               mac_string (str): a readable string (e.g. '01:02:03:04:05:06')
           Returns:
               str: a MAC address in hex form
        """
    sp = mac_string.split(':')
    mac_string = ''.join(sp)
    return binascii.unhexlify(mac_string)


def inet_to_str(inet):
    """Convert inet object to a string

        Args:
            inet (inet struct): inet network address
        Returns:
            str: Printable/readable IP address
    """
    # First try ipv4 and then ipv6
    try:
        return socket.inet_ntop(socket.AF_INET, inet)
    except ValueError:
        return socket.inet_ntop(socket.AF_INET6, inet)


def str_to_inet(address):
    """Convert an a string IP address to a inet struct

        Args:
            address (str): String representation of address
        Returns:
            inet: Inet network address
    """
    # First try ipv4 and then ipv6
    try:
        return socket.inet_pton(socket.AF_INET, address)
    except socket.error:
        return socket.inet_pton(socket.AF_INET6, address)


def is_internal(ip_address):
    """Determine if the address is an internal ip address"""
    return ipaddress.ip_address(ip_address).is_private


def is_special(ip_address):
    """Determine if the address is SPECIAL
       Note: This is super bad, improve it
    """
    special = {'224.0.0.251': 'multicast_dns',
               'ff02::fb': 'multicast_dns'}
    return special[ip_address] if ip_address in special else False


def traffic_direction(conn_row):
    """Determine the direction of the connection traffic (takes a conn.log row)"""

    # First try to use the local orig/resp fields
    if conn_row.get('local_orig') and conn_row.get('local_resp'):
        local_orig = conn_row['local_orig']
        local_resp = conn_row['local_resp']
    else:
        # Well we don't have local orig/resp fields so use RFC1918 logic
        local_orig = ipaddress.ip_address(conn_row['id.orig_h']).is_private
        local_resp = ipaddress.ip_address(conn_row['id.resp_h']).is_private

    # Determine north/south or internal traffic
    if (not local_orig) and local_resp:
        return 'incoming'
    if local_orig and not local_resp:
        return 'outgoing'

    # Neither host is in the allocated private ranges
    if ipaddress.ip_address(conn_row['id.orig_h']).is_multicast or \
       ipaddress.ip_address(conn_row['id.resp_h']).is_multicast:
        return 'multicast'

    # Both hosts are internal
    return 'internal'


def test_utils():
    """Test the utility methods"""

    print(mac_to_str(b'\x01\x02\x03\x04\x05\x06'))
    assert mac_to_str(b'\x01\x02\x03\x04\x05\x06') == '01:02:03:04:05:06'
    assert str_to_mac('01:02:03:04:05:06') == b'\x01\x02\x03\x04\x05\x06'
    my_mac = b'\x01\x02\x03\x04\x05\x06'
    my_str = mac_to_str(my_mac)
    assert str_to_mac(my_str) == my_mac
    print(inet_to_str(b'\x91\xfe\xa0\xed'))
    assert inet_to_str(b'\x91\xfe\xa0\xed') == '145.254.160.237'
    assert str_to_inet('145.254.160.237') == b'\x91\xfe\xa0\xed'

    # IPV6 for Google DNS server
    google_dns = '2001:4860:4860::8888'
    google_inet = b' \x01H`H`\x00\x00\x00\x00\x00\x00\x00\x00\x88\x88'
    assert str_to_inet(google_dns) == google_inet

    # Various methods
    assert is_internal('10.0.0.1')
    assert is_internal('222.2.2.2') is False
    assert is_special('224.0.0.251')
    assert is_special('224.0.0.252') is False
    print('Success!')


if __name__ == '__main__':
    test_utils()



"""Cert Checker zat Example"""

import os
import sys
import argparse
from pprint import pprint

# Local imports
from zat import zeek_log_reader
from zat.utils import vt_query

if __name__ == '__main__':
    # Example to check all the x509 Certs from 'Let's Encrypt' for potential phishing/malicious sites

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('zeek_log', type=str, help='Specify a zeek log to run ZeekLogReader test on')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Sanity check that this is a dns log
    if 'x509' not in args.zeek_log:
        print('This example only works with Zeek x509.log files..')
        sys.exit(1)

    # File may have a tilde in it
    if args.zeek_log:
        args.zeek_log = os.path.expanduser(args.zeek_log)

        # Create a VirusTotal Query Class
        vtq = vt_query.VTQuery()

        # These domains may be spoofed with a certificate issued by 'Let's Encrypt'
        spoofed_domains = set(['paypal', 'gmail', 'google', 'apple','ebay', 'amazon'])

        # Run the zeek reader on the x509.log file looking for spoofed domains
        reader = zeek_log_reader.ZeekLogReader(args.zeek_log, tail=True)
        for row in reader.readrows():

            # Pull out the Certificate Issuer
            issuer = row['certificate.issuer']
            if "Let's Encrypt" in issuer:

                # Check if the certificate subject has any spoofed domains
                subject = row['certificate.subject']
                if any([domain in subject for domain in spoofed_domains]):
                    print('\n<<< Suspicious Certificate Found >>>')
                    pprint(row)

                    # Make a Virus Total query with the spoofed domain (just for fun)
                    query_domain = subject[3:] # Just chopping off the 'CN=' part
                    results = vtq.query_url(query_domain)
                    if results.get('positives', 0) >= 2: # At least two hits
                        print('\n<<< Virus Total Query >>>')
                        pprint(results)



"""Read Kafka Streams and Print out the messages"""

import sys
import argparse
from pprint import pprint
import json
from kafka import KafkaConsumer
from kafka.errors import NoBrokersAvailable

# Local imports
from zat.utils import signal_utils


def exit_program():
    """Exit on Signal"""
    print('Exiting Program...')
    sys.exit()


if __name__ == '__main__':
    """Read Kafka Streams and Print out the messages"""

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('--server', type=str, default='localhost:9092',
                        help='Specify the Kafka Server (default: localhost:9092)')
    parser.add_argument('--topics', type=lambda s: s.split(','), default='all',
                        help='Specify the Kafka Topics (e.g. dns   or   dns, http, blah   (defaults to all)')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Create a Kafka Consumer and subscribe to the topics
    all_topics = ['capture_loss', 'dns', 'http', 'ssl', 'weird', 'conn', 'files', 'x509']
    kserver = args.server
    topics = args.topics if args.topics != ['all'] else all_topics
    print('Subscribing to: {!r}'.format(topics))
    try:
        consumer = KafkaConsumer(*topics, bootstrap_servers=[kserver],
                                 value_deserializer=lambda x: json.loads(x.decode('utf-8')))
    except NoBrokersAvailable:
        print('Could not connect to Kafka server: {:s}'.format(args.server))
        sys.exit(-1)

    # Launch long lived process with signal catcher
    with signal_utils.signal_catcher(exit_program):

        # Now lets process our Kafka Messages
        for message in consumer:
            topic = message.topic
            message = message.value
            print('\n{:s}'.format(topic.upper()))
            pprint(message)
